;;; $Id: kScheme1.scm,v 1.3 1996/09/21 10:10:18 queinnec Exp $

(define (start-scheduler thunk)
  (add-to-scheduler! thunk)
  (call/cc (lambda (return)
             (set! *return* return)
             (call-scheduler!) )) )

(define (add-to-scheduler! f . args)
  (set! *threads* (append *threads* (list (cons f args)))) )

(define (call-scheduler!)
  (if (pair? *threads*)
      (let ((thread (car *threads*)))
        (set! *threads* (cdr *threads*))
        (apply (car thread) (cdr thread)) )
      (*return* "No more threads to run") ) )

(define *threads* '())

(define *return* (lambda (x) x))

;;; An evaluate stuffed with thread creations.

(define (evaluate e r k)
  (if (atom? e)
      (cond ((symbol? e) (add-to-scheduler! k (lookup e r)))
            (else (add-to-scheduler! k e)) )
      (let ((special (assq (car e) *specials*)))
        (if (pair? special)
            (add-to-scheduler! (cdr special) e r k)
            (add-to-scheduler! evaluate-application (car e) (cdr e) r k) ) ) )
  (call-scheduler!) )

(defprimitive suicide
  (lambda ()
    (call-scheduler!) )
  0 )

(define-special ((parallel . forms) r k)
  (for-each (lambda (form)
              (add-to-scheduler! evaluate form r k) )
            forms )
  (call-scheduler!) )

(define-special ((future form) r k)
  (let ((box (list '*future* 'wait '())))
    ;; marquer que la bo\^ite est non initialis\'ee.
    (set-car! (cdr box) box)
    (add-to-scheduler! evaluate 
                       form 
                       r
                       (lambda (v)
                         (set-car! (cdr box) v)
                         (for-each (lambda (task)
                                     (add-to-scheduler! task v) )
                                   (caddr box) )
                         (call-scheduler!) ) )
    (add-to-scheduler! k box)
    (call-scheduler!) ) )

(definitial touch
  (lambda (k args)
    (if (= 1 (length args))
        (let ((box (car args)))
          (if (and (pair? box) (eq? '*future* (car box)))
              (begin
                (if (eq? box (cadr box))
                    (set-car! (cddr box) (cons k (caddr box)))
                    (add-to-scheduler! k (cadr box)) )
                (call-scheduler!) )
              (wrong "Not a future" box) ) )
        (wrong "Incorrect arity for" 'touch) ) ) )

(define-special ((qlambda variables . body) r k)
  (k (make-qlambda-function variables body r)) )

(define (make-qlambda-function variables body r)
  (let ((active? #f)
        (queue   '()) )
    (lambda (k arguments)
      (if active?
          (set! queue (append queue (list (cons k arguments))))
          (let enter ((k k) (arguments arguments))
            (set! active? #t)
            (add-to-scheduler! evaluate-begin 
                               body
                               (extend r variables arguments)
                               (let ((exited? #f))
                                 (lambda (v)
                                   (if (not exited?)
                                       (if (pair? queue)
                                           (let ((k+args (car queue)))
                                             (set! queue (cdr queue))
                                             (add-to-scheduler!
                                              enter (car k+args) 
                                                    (cdr k+args) ) )
                                           (set! active? #f) ) )
                                   (set! exited? #t)
                                   (k v) ) ) ) ) )
      (call-scheduler!) ) ) )

(definitial sleep
  (lambda (k args)
    (if (= 1 (length args))
        (let wait ((count (car args)))
          (if (<= count 0)
              (add-to-scheduler! k count)
              (add-to-scheduler! (lambda ()
                                   (wait (- count 1)) )) )
          (call-scheduler!) )
        (wrong "Incorrect arity for" 'sleep) ) ) )

;;; Channels.

(defprimitive channel-create
  (lambda ()
    (vector '*channel* '() '()) )
  0 )

(define (Channel? o)
  (and (vector? o)
       (eq? '*channel* (vector-ref o 0)) ) )
(define (Channel-values c)
  (vector-ref c 1) )
(define (set-Channel-values! c v)
  (vector-set! c 1 v) )
(define (Channel-pending c)
  (vector-ref c 2) )
(define (set-Channel-pending! c v)
  (vector-set! c 2 v) )

(defprimitive channel-send
  (lambda (c v)
    (if (Channel? c)
        (if (pair? (Channel-pending c))
            (let ((k (car (Channel-pending c))))
              (set-Channel-pending! c (cdr (Channel-pending c)))
              (add-to-scheduler! k v) )
            (begin
              (set-Channel-values! c (append (Channel-values c) (list v)))
              nothing-particular ) )
        (wrong "Not a channel" c) ) )
  2 )

(definitial channel-receive
  (lambda (k args)
    (if (= 1 (length args))
        (let ((c (car args)))
          (if (Channel? c)
              (begin 
                (if (pair? (Channel-values c))
                    (let ((v (car (Channel-values c))))
                      (set-Channel-values! c (cdr (Channel-values c)))
                      (add-to-scheduler! k v) )
                    (set-Channel-pending! 
                     c (append (Channel-pending c) (list k)) ) )
                (call-scheduler!) )
              (wrong "Not a channel" c) ) )
        (wrong "Incorrect arity for" 'channel-receive) ) ) )

;;; Toplevel must now collect all results.

(define (kscheme1 . e)
  (define (toplevel e)
    (let ((results '()))
      (start-scheduler
       (lambda ()
         (evaluate e
                   r.init
                   (lambda (v)
                     (display "kScheme= ")(display v)(newline)
                     (set! results (cons v results))
                     (call-scheduler!) ) )
         results ) ) ) )
  (if (pair? e)
      (toplevel (car e))
      (interpreter
       "kScheme? " "" #t
       (lambda (read print error)
         (set! wrong error)
         (lambda ()
           (toplevel (read))
           (print "") ) ) ) ) )

(define (test-kscheme1 file)
  (suite-test file "kScheme? " "" #t
   (lambda (read check error)
     (set! wrong error)
     (lambda () 
       (let ((results '()))
         (start-scheduler
          (lambda ()
            (evaluate (read) 
                      r.init 
                      (lambda (v) 
                        (display "kScheme= ")(display v)(newline)
                        (set! results (cons v results)) 
                        (call-scheduler!) ) ) ) )
         (check results) ) ) )
   check-kscheme-result ) )

(define (check-kscheme-result expected obtained)
  (or (and (pair? expected)
           (case (car expected)
             ((multiple)
              (set-equal? (cdr expected) obtained) )
             ((or)
              (and (pair? (cdr expected))
                   (or (check-result (cadr expected) obtained)
                       (check-result `(or . ,(cddr expected)) obtained) ) ) )
             (else (naive-match expected obtained)) ) )
      (and (pair? obtained)
           (null? (cdr obtained))
           (naive-match expected (car obtained)) ) ) )

;;; end of kScheme1.scm
