;;; $Id: natScheme.scm,v 1.6 1996/09/25 10:49:30 queinnec Exp $
;;; snarfed from ~/DEA/natScheme.scm

;;; A naive Scheme interpreter in Scheme.
;;; Christian Queinnec
;;; Universite Paris 6 & INRIA-Rocquencourt

(define (atom? e) 
  (not (pair? e)) )

(define (evaluate e env)
  (if (atom? e)  ; \ie {\tt (not (pair? e))}
      (cond ((symbol? e) (lookup e env))
            ((or (number? e) (string? e) (char? e) (boolean? e)) 
             e )
            (else (wrong "Cannot evaluate" e)) )
      (case (car e)
        ;; pas de v\'erification syntaxique
        ((quote) (cadr e))
        ((if) (if (evaluate (cadr e) env)
                  (evaluate (caddr e) env)
                  (evaluate (cadddr e) env) ))
        ;;\ {\tt begin} n'est pas une vraie forme sp\'eciale
        ((begin) (eprogn (cdr e) env))
        ((set!) (update! (cadr e) env (evaluate (caddr e) env)))
        ((lambda) (make-function (cadr e) (cddr e) env))
        (else (invoke (evaluate (car e) env)
                      (evlis (cdr e) env) )) ) ) )

;;; Two specialized evaluators.

(define (evlis exps env)
  (if (pair? exps)
      (cons (evaluate (car exps) env)
            (evlis (cdr exps) env) )
      '() ) )

(define (eprogn exps env)
  (if (pair? exps)
      (if (pair? (cdr exps))
          (begin (evaluate (car exps) env)
                 (eprogn (cdr exps) env) )
          (evaluate (car exps) env) )
      '() ) )

;;; Functions are represented as functions

(define (make-function variables body env)
  (lambda (values)
     (eprogn body (extend env variables values)) ) )

(define (invoke fn args)
  (if (procedure? fn) 
      (fn args)
      (wrong "Not a function" fn) ) )

;;; Environments are Alists handled by lookup and update!.
;;; The global environment is not exensible (see definitial below).

(define (lookup id env)
  (if (pair? env)
      (if (eq? (caar env) id)
          (cdar env)
          (lookup id (cdr env)) )
      (wrong "No such binding" id) ) )

(define (update! id env value)
  (if (pair? env)
      (if (eq? (caar env) id)
          (set-cdr! (car env) value)
          (update! id (cdr env) value) )
      (wrong "No such binding" id) ) )

(define (extend env names values)
  (cond ((pair? names)
         (if (pair? values)
             (cons (cons (car names) (car values))
                   (extend env (cdr names) (cdr values)) )
             (wrong "Too less values") ) )
        ((null? names)
             (if (null? values)
                 env 
                 (wrong "Too much values") ) )
        ((symbol? names) (cons (cons names values) env)) ) )

(define env.init '())

;;; Three macros to define the initial global environment

(define env.global env.init)

(define-macro (definitial name value)
  `(begin (set! env.global (cons (cons ',name ,value) env.global))
          ',name ) )

(define-macro (defprimitive name value arity)
  `(definitial ,name 
     (lambda (values) 
       (if (= (length values) ,arity)
           (apply ,value values)
           (wrong "Incorrect arity"
                  (list ',name values) ) ) ) ) )

(define-macro (defnaryprimitive name value arity)
  `(definitial ,name 
     (lambda (values) 
       (if (>= (length values) ,arity)
           (apply ,value values)
           (wrong "Incorrect arity"
                  (list ',name values) ) ) ) ) )

(define the-false-value #f)

(define-macro (defpredicate name value arity)
  `(definitial ,name
     (lambda (values) 
       (if (= ,arity (length values))
           (or (apply ,value values) the-false-value)
           (wrong "Incorrect arity"
                  (list ',name values) ) ) ) ) )

(defprimitive cons cons 2)
(defprimitive car car 1)
(defprimitive cdr cdr 1)
(defpredicate pair? pair? 1)
(defpredicate symbol? symbol? 1)
(defpredicate eq? eq? 2)
(defprimitive set-car! set-car! 2)
(defprimitive set-cdr! set-cdr! 2)
(defnaryprimitive + + 1)
(defprimitive - - 2 )
(defprimitive = = 2)
(defpredicate < < 2)
(defpredicate > > 2)
(defnaryprimitive * * 1)
(defpredicate <= <= 2)
(defpredicate >= >= 2)
(defprimitive remainder remainder 2)

(defprimitive call/cc 
   (lambda (f) 
     (call/cc (lambda (g) 
                (invoke
                 f (list (lambda (values)
                           (if (= (length values) 1)
                               (g (car values))
                               (wrong "Incorrect arity" g) ) ))) )) )
   1 )

(defnaryprimitive apply 
  (lambda values
    (define (listify arguments)
      (if (pair? (cdr arguments))
          (cons (car arguments) (listify (cdr arguments)))
          (car arguments) ) )
    (invoke (car values) (listify (cdr values))) )
  2 )

;;; Some fundamental constants:
(definitial t #t)
(definitial f #f)
(definitial nil '())

(defnaryprimitive list
   (lambda values values)
   0 ) 

;;; Preserve the common environment.

(define env.predefined env.global)

;;; Some free global locations:
(definitial x 'void)
(definitial y 'void)
(definitial z 'void)
(definitial a 'void)
(definitial b 'void)
(definitial c 'void)
(definitial foo 'void)
(definitial bar 'void)
(definitial hux 'void)
(definitial fib 'void)
(definitial fact 'void)
(definitial visit 'void)
(definitial length 'void)
(definitial primes 'void)

;;; Starting the interpreter.

(define (toplevel)
  (display (evaluate (read) env))
  (toplevel env.global) )

(define (scheme)
  (display "Bienvenue !")
  (newline)
  (let toplevel ()
    (display "?? ")
    (let ((e (read)))
      (if (eof-object? e)
          (begin (display ";;; end")
                 (newline) )
          (let ((r (evaluate e env.global)))
            (display "== ")
            (display r)
            (newline)
            (toplevel) ) ) ) ) )

;;; Testing
;;; These tests require ../DEA/tester.scm

(define (natscheme)
  (interpreter "NatScheme? "  "NatScheme= " #t
    (lambda (read print error)
      (set! wrong error)
      (lambda ()
         (print (evaluate (read) env.global)) ) ) ) )

(define tests-to-skip '())

(define (test-natscheme file)
  (suite-test file "NatScheme? " "NatScheme= " 
     #t
     (lambda (read check error)
       (set! wrong error)
       (lambda ()   
         (define (skip-read)
           (let ((e (read)))
             (if (member e tests-to-skip)
                 (begin (read)              ; skip the associated result
                        (skip-read) )
                 e ) ) )
         (check (evaluate (skip-read) env.global) ) ) )
     equal? ) )
  
;(display "Run the interpreter with (natscheme)
;and test suite with (test-natscheme #t)
;" )

;;; end of natScheme.scm
